@@ -1,3 +1,5 @@
+#![allow(clippy::uninlined_format_args)]
+
 #[cfg(not(any(feature = "bat-printer", feature = "syntect-printer")))]
 compile_error!("Either feature \"bat-printer\" or \"syntect-printer\" must be enabled");
 
@@ -5,7 +7,7 @@ pub mod chunk;
 pub mod grep;
 pub mod printer;
 
-mod io;
+mod broken_pipe;
 
 #[cfg(feature = "bat-printer")]
 pub mod bat;
